# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Scmbug::Daemon::Integration;


use strict;
use Data::Dumper;
use Scmbug::Daemon::Request;
use Scmbug::Common qw (
				     array_ref_is_in_set

				     $TAG_OPERATION_ADD
				     $TAG_OPERATION_DELETE
				     $TAG_OPERATION_MOVE
				     );

# Constructor
sub new {
    my $type = shift;
    my $data = shift;
    my $self = {};

    bless $self, $type;
    if ( !defined( $data ) ) {
	print STDERR "No bugtracker set in Integration->new()\n";
	exit -1;
    }
    $self->bugtracker( $data );

    return $self;
}



sub bugtracker {
    my $self = shift;
    my $data = shift;

    if ( $data ) {
	$self->{ bugtracker } = $data;
    } else {
	return $self->{ bugtracker };
    }
}





#
# Processes a commit activity
#
sub process_activity_commit {
    my $self = shift;
    my $request = shift;
    my $retval;

    $self->bugtracker()->integration_connect_database();

    # Add integration comments in the bug-tracker
    $retval = $self->add_comment_to_all( $request,
					 $request->{ translated_username },
					 $request->{ log_message },
					 $request->{ ids } );

    if ( $retval != 0 ) {
	# There was a problem adding integration comments. Thus we
	#w on't attempt to change the bug resolution at all.
    } else {
	# Change the bug resolution
	$retval = $self->change_resolution_in_all ( $request,
						    $request->{ translated_username } );
    }

    $self->bugtracker()->integration_disconnect_database();

    return $retval;
}



#
# Processes a verify activity
#
sub process_activity_verify {
    my $self = shift;
    my $request = shift;
    my $ret_message = "";

    $self->bugtracker()->integration_connect_database();

    # Perform various verification tests for all these bugs
    foreach my $id ( @{ $request->{ ids } } ) {
	my $bugid = $id;
	my $who = $request->{ translated_username };
	my $ret;
	my $bug_state;

	# Apply policy valid_product_name
	if ( $request->{ policies }->{ valid_product_name }->{ enabled } ) {
	    my $bug_product_name;
	    $bug_product_name = $self->bugtracker()->integration_get_product_name( $bugid );
	    if ( $bug_product_name ne $request->{ product_name } ) {
		
		# This bug does not belong to the list of products
		# specified by the SCM system. Report an error accordingly.
		$ret_message = "Bug '$bugid' does not belong to the product '" . $request->{ product_name } . "'";
		$ret_message .= ". It belongs to product '" . $bug_product_name . "'.\n";
		$request->add_result(1, $ret_message );
	    }
	}

	# Apply policy valid_bug_owner
	if ( $request->{ policies }->{ valid_bug_owner }->{ enabled } ) {
	    # Verify that a bug is owned by the user submitting the
	    # request
	    my @bug_owners = $self->bugtracker()->integration_get_bug_owners( $bugid );
            my ( $matched_owner, $case_sensitive_username_verification_who ) = $self->verify_bug_ownership( $who, $request->{ policies }->{ case_sensitive_username_verification }, @bug_owners );

            if( $matched_owner == 0 ) {
                my $owners_string = join( ',', @bug_owners );
                $ret_message = "Bug '$bugid' is not assigned to '$case_sensitive_username_verification_who'. It is assigned to '$owners_string'.\n";
                $request->add_result(1, $ret_message);
            }
	}

	# Apply policy open_bug_state
	if ( $request->{ policies }->{ open_bug_state }->{ enabled } ) {
	    # Verify that the bug against which the activity is issued
	    # is in an open bug state
	    ($ret, $bug_state) = $self->bugtracker()->integration_bug_in_active_state( $bugid );
	    if ( $ret == 0) {
		$ret_message = "Bug '$bugid' is in the '$bug_state' state, instead of the " . $self->bugtracker()->integration_active_states_list() . ".\n";
		$request->add_result(1, $ret_message);
	    }
	}

	# Check if the bug is locked by the bug-tracker. This is
	# required because the bug-tracker may acquire a write lock on
	# given activities (e.g. backups)
        my $lock_available = $self->bugtracker()->integration_bug_lock_available( $id );

	if( $lock_available == 0 ) {
	    $ret_message = "Bug '$id' is currently locked.\n";
	    $request->add_result(1, $ret_message);
	}

    }



    #
    # Apply policy resolution_template
    #
    if ( $request->{ policies }->{ resolution_template }->{ enabled } ) {

	foreach my $resolution_id ( sort keys %{ $request->{ resolution_ids } } ) {
	    my $who = $request->{ translated_username };

	    # Verify that the bugs are owned by the user changing the
	    # resolution status
	    if ( $request->{ policies }->{ resolution_template }->{ resolution_valid_product_name } ) {
		my $bug_product_name;
		$bug_product_name = $self->bugtracker()->integration_get_product_name( $resolution_id );
		if ( $bug_product_name ne $request->{ product_name } ) {
		    
		    # This bug does not belong to the list of products
		    # specified by the SCM system. Report an error accordingly.
		    $ret_message = "Bug '$resolution_id' used for bug resolution does not belong to the product '" . $request->{ product_name } . "'";
		    $ret_message .= ". It belongs to product '" . $bug_product_name . "'.\n";
		    $request->add_result(1, $ret_message );
		}
	    }

	    # Verify that the bugs are owned by the user changing the
	    # resolution status
	    if ( $request->{ policies }->{ resolution_template }->{ resolution_valid_bug_owner } ) {
		my @bug_owners = $self->bugtracker()->integration_get_bug_owners( $resolution_id );
                my ( $matched_owner, $case_sensitive_username_verification_who ) = $self->verify_bug_ownership( $who, $request->{ policies }->{ case_sensitive_username_verification }, @bug_owners );
                if( $matched_owner == 0 ) {
                    my $owners_string = join( ',', @bug_owners );
                    $ret_message = "Bug '$resolution_id' used for bug resolution is not assigned to '$case_sensitive_username_verification_who'. It is assigned to '$owners_string'.\n";
                    $request->add_result(1, $ret_message);
                }
	    }

	    # Verify that the resolution status change is valid
	    my ( $resolution_change_retval, $current_resolution_state ) = $self->bugtracker()->integration_bug_resolution_change_is_valid( $resolution_id, $request->{ resolution_ids }->{ $resolution_id }->{ status }, $request->{ resolution_ids }->{ $resolution_id }->{ resolution }, $request->{ policies }->{ resolution_status_case_sensitive_verification } );
	    if ( $resolution_change_retval == 1 ) {
		$ret_message = "Resolving bug '$resolution_id' from state '$current_resolution_state' to state '" . $request->{ resolution_ids }->{ $resolution_id }->{ status } . "' would violate the workflow of the bug-tracker. Instead, the resolution could be changed to the " . $self->bugtracker()->integration_valid_resolution_states_list( $current_resolution_state ) . ".\n";
		$request->add_result(1, $ret_message);
	    } elsif ( $resolution_change_retval == 2 ) {
		$ret_message = "Invalid resolution '" . $request->{ resolution_ids }->{ $resolution_id }->{ resolution } . "' when resolving bug '$resolution_id' from state '$current_resolution_state' to state '" . $request->{ resolution_ids }->{ $resolution_id }->{ status } . "'. Instead, a valid resolution could be the " . $self->bugtracker()->integration_valid_resolutions_list( $request->{ resolution_ids }->{ $resolution_id }->{ status }, $request->{ policies }->{ resolution_status_case_sensitive_verification } ) . ".\n";
		$request->add_result(1, $ret_message);
	    }
	
	    # Check if the bug is locked by the bug-tracker. This is
	    # required because the bug-tracker may acquire a write lock on
	    # given activities (e.g. backups)
	    my $lock_available = $self->bugtracker()->integration_bug_lock_available( $resolution_id );

	    if( $lock_available == 0 ) {
		$ret_message = "Bug '$resolution_id' is currently locked.\n";
		$request->add_result(1, $ret_message);
	    }
	}
    }

    $self->bugtracker()->integration_disconnect_database();
}


#
# Processes a tagging activity. Assumes only one product name was
# specified
#
# RETURNS:
# - 0 if the tag already exists.
# - 1 otherwise
sub process_activity_tag {
    my $self = shift;
    my $request = shift;
    my $ret_message;

    $self->bugtracker()->integration_connect_database();

    if ( $request->{ type }->{ tag }->{ operation } eq $TAG_OPERATION_ADD ) {
	# Add the tag if it's not already there
        my $ret = $self->bugtracker()->integration_tag_exists( $request->{ product_name },
                                                               $request->{ type }->{ tag }->{ name } );
	if ( $ret == 0) {
	    # This tag does not exist already for this product. Add it
	    $ret = $self->bugtracker()->integration_add_tag( $request->{ product_name },
							     $request->{ type }->{ tag }->{ name },
							     $request->{ original_log_message } );
	    
	    if ( $ret == 1 ) {
		$ret_message = "Tag '" . $request->{ type }->{ tag }->{ name } . 
		    "' could not be added on product '" . $request->{ product_name } . "' because no product with this name exists.\n";
		$request->add_result( 1, $ret_message );
	    }
	} elsif ( $ret == 1 ) {
            # The tag exists.
            #
	    # An SCM system mistakenly issued multiple tag activities
	    # (e.g. CVS). Do not add the tag again.
	    $self->bugtracker()->integration_disconnect_database();
	    return 0;
	} elsif ( $ret == 2 ) {
            # The product name does not exist 
            $ret_message = "Tag '" . $request->{ type }->{ tag }->{ name } . 
                "' could not be added on product '" . $request->{ product_name } . "' because no product with this name exists.\n";
            $request->add_result( 1, $ret_message );
	}
    } elsif ( $request->{ type }->{ tag }->{ operation } eq $TAG_OPERATION_DELETE ) {
	# Delete the tag

        # We should be checking if the tag exists, and failing if it
        # doesn't. But an SCM system might have mistakenly issued
        # multiple tag activities (e.g. CVS). Do not report an error
        # the second time the tag is requested to be deleted again.
	my $ret;
	$ret = $self->bugtracker()->integration_delete_tag( $request->{ product_name },
							    $request->{ type }->{ tag }->{ name } );
	
	if ( $ret == 1) {
	    $ret_message = "Tag '" . $request->{ type }->{ tag }->{ name } . 
		"' could not be deleted from product '" . $request->{ product_name } . "' because no product with this name exists.\n";
	    $request->add_result( 1, $ret_message );
	}
    } elsif ( $request->{ type }->{ tag }->{ operation } eq $TAG_OPERATION_MOVE ) {
	# Do nothing
    }

    $self->bugtracker()->integration_disconnect_database();

    return 1;
}



#
# Processes a VDD Query. Assumes only one product name was
# specified
#
sub process_activity_get_vdd {
    my $self = shift;
    my $request = shift;
    my $vdd;

    $self->bugtracker()->integration_connect_database();
    $vdd = $self->bugtracker()->integration_get_vdd( $request->{ product_name },
						     $request->{ type }->{ get_vdd }->{ from_version },
						     $request->{ type }->{ get_vdd }->{ to_version },
						     $request->{ type }->{ get_vdd }->{ from_label_date },
						     $request->{ type }->{ get_vdd }->{ to_label_date } );
    $self->bugtracker()->integration_disconnect_database();

    return $vdd;
}



#
# Processes a VDD Query. Assumes only one product name was
# specified
#
sub process_activity_get_bugs {
    my $self = shift;
    my $request = shift;
    my $mergelist;
    my $ret_val;
    my $ret_message;

    $self->bugtracker()->integration_connect_database();
    ($mergelist, $ret_val, $ret_message) = $self->bugtracker()->integration_get_bugs( $request->{ product_name },
										      $request->{ ids } );
    $self->bugtracker()->integration_disconnect_database();


    if ($ret_val == 1) {
	$request->add_result( 1, $ret_message );
    }

    return $mergelist;
}



# Adds a comment from a user to a list of bugs
#
# RETURNS:
# - 0 on success
# - 1 if an empty list of bug ids was supplied
# - 2 The comment could not be added
sub add_comment_to_all {
    my $self = shift;
    my ( $request, $username, $comment, $bugs_list ) = ( @_ );
    my $appended = 0;
    my $comment_addition_retval = 0;
    my $comment_addition_retval_string = "";

    # For each bug id, append a comment to the bug
    foreach my $id ( @{ $bugs_list } ) {
	( $comment_addition_retval, $comment_addition_retval_string ) = $self->bugtracker()->integration_add_comment( $id, $username, $comment );
	$appended = 1;
    }

    if ( $comment_addition_retval == 1 ) {
	# Adding the comment failed in the bugtracker backend, most
	# likely due to a bug in the bugtracker
	$request->add_result( 1, $comment_addition_retval_string );
	return 2;
    } else {
	if ( $appended ) {
	    return 0;
	} else {
	    # The list of bug ids was empty.
	    $request->add_result( 1, "An empty list of bug ids was supplied. This should have not been permitted by check_request() in the first place\n" );
	    return 1;
	}
    }
}


# Changes the bug resolution of a list of bugs
#
# RETURNS:
# - 0 on success
# - 1 on failure
sub change_resolution_in_all {
    my $self = shift;
    my ( $request, $username ) = ( @_ );

    # Only change bug resolutions if it's been requested
    if ( $request->{ policies }->{ resolution_template }->{ enabled } ) {

	foreach my $resolution_id ( sort keys %{ $request->{ resolution_ids } } ) {
	    my ( $retval, $retval_string ) = $self->bugtracker()->integration_change_bug_resolution( $resolution_id, $username, $request->{ resolution_ids }->{ $resolution_id }->{ status }, $request->{ resolution_ids }->{ $resolution_id }->{ resolution }, $request->{ resolution_ids }->{ $resolution_id }->{ resolution_data } );
	    if ( $retval != 0) {
		$request->add_result( 1, $retval_string );
		return 1;
	    }
	}
    }

    return 0;
}



# Get a list of email address to which mail should be sent as a result
# of this request
#
# NOTE:
# - $bugid is an optional parameter. It is not supplied when no bug
#   ids were defined
# - Email should always be replied to the entire From list. Thus,
#   every time an address is set in 'From', it is also set in
#   'Reply-to'.
#
# RETURNS:
# - A hash containing as keys the list of email addresses
sub get_mailing_list {
    my $self = shift;
    my $request = shift;
    my $mailing_list_for_a_tag = shift;
    my $bugid = shift;
    my $mailing_list;


    # Get the email address of the SCM user, regardless. This email
    # will be used as one of the addresses in the 'Reply-to' header.
    my $scm_user_email = $self->bugtracker()->integration_get_user_email( $request->{ translated_username } );
    if ($request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_scm_user } == 1) {
        $mailing_list->{ To }->{ $scm_user_email } = 1;
    }

    # Get the email address of the bug owner
    if ($request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_bug_owner } == 1) {
	my $bug_owner_email;

	$bug_owner_email = $self->bugtracker()->integration_get_user_email( $request->{ translated_username } );
        $mailing_list->{ To }->{ $bug_owner_email } = 1;
    }

    # Get the email address of the user which reported the bug
    if ($request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_bug_reporter } == 1) {
	my $bug_reporter_email;

	if ( defined($bugid) ) {
	    $bug_reporter_email = $self->bugtracker()->integration_get_bug_reporter_email( $bugid );
	    $mailing_list->{ To }->{ $bug_reporter_email } = 1;
	}
    }

    # Get the email addresses of all users who monitor the bug
    if ($request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_bug_monitors } == 1) {
	my @bug_monitors_emails;
	
	if ( defined($bugid) ) {
	    @bug_monitors_emails = $self->bugtracker()->integration_get_bug_monitors_email_list( $bugid );
	    foreach my $bug_monitor ( @bug_monitors_emails ) {
		$mailing_list->{ To }->{ $bug_monitor } = 1;
	    }
	}
    }

    # If we are tagging a product, also notify the product's owners
    if ($mailing_list_for_a_tag == 1){
	# Get the email addresses of the product owners (managers)
	if ($request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_product_owners } == 1) {
	    my @product_owners_emails;
	    
	    @product_owners_emails = $self->bugtracker()->integration_get_managers_email_list( $request->{ product_name } );
	    
	    foreach my $product_owner ( @product_owners_emails ) {
		$mailing_list->{ To }->{ $product_owner } = 1;
	    }
	}
    }

    # Set the remaining of the recipient list
    if ( $request->{ policies }->{ mail_notification }->{ mail_settings }->{ To } ne '' ) {
	$mailing_list->{ To }->{ $request->{ policies }->{ mail_notification }->{ mail_settings }->{ To } } = 1;
    }


    # Add the Sender.
    if ( $request->{ policies }->{ mail_notification }->{ mail_settings }->{ From } ne '' ) {
	$mailing_list->{ From }->{ $request->{ policies }->{ mail_notification }->{ mail_settings }->{ From } } = 1;
	$mailing_list->{ 'Reply-To' }->{ $request->{ policies }->{ mail_notification }->{ mail_settings }->{ From } } = 1;
    }

    # If the email should also appear to have been sent from the SCM
    # user, add that user first in the From list. He is added at a
    # higher priority (value 0)
    if ( $request->{ policies }->{ mail_notification }->{ mail_recipients }->{ mail_also_appears_from_scm_user } == 1) {
	$mailing_list->{ From }->{ $scm_user_email } = 0;
	$mailing_list->{ 'Reply-To' }->{ $scm_user_email } = 0;
    } else {
	# But if the SCM user was not set in From, it should also be
	# replied to the SCM user first.
	$mailing_list->{ 'Reply-To' }->{ $scm_user_email } = 0;
    }

    return $mailing_list;
}



# Verify that a user is among the list of owners of a bug
#
# PARAMETERS
# $1 - The username that will be verified
# $2 - A flag indicating if case sensitive verification will be
#      applied
# $3 - An array holding the username of the bug owners
#
# RETURNS:
# - 1 or 0 if the user is among the list of owners, <the case
# sensitive username used for the verification of the user>
sub verify_bug_ownership {
    my $self = shift;
    my $who = shift;
    my $case_sensitive_username_verification = shift;
    my @bug_owners = ( @_ );

    my $case_sensitive_username_verification_who;
    if ( $case_sensitive_username_verification == 1 ) {
        $case_sensitive_username_verification_who = $who;
    } else {
        $case_sensitive_username_verification_who = lc $who;
    }

    my $matched_owner = 0;
    foreach my $bug_owner ( @bug_owners ) {
        # Apply policy case_sensitive_username_verification
        my $case_sensitive_username_verification_bug_owner;
        if ( $case_sensitive_username_verification == 1 ) {
            $case_sensitive_username_verification_bug_owner = $bug_owner;
        } else {
            $case_sensitive_username_verification_bug_owner = lc $bug_owner;
        }
        
        if ( $case_sensitive_username_verification_bug_owner eq $case_sensitive_username_verification_who ) {
            $matched_owner = 1;
        }
    }

    return ( $matched_owner, $case_sensitive_username_verification_who );
}



1;
